

/* Coyright (c) 2004-2005  Petr Vandrovec.   Licensed under LGPL-2.1
   2006 Kjetil Matheussen.
 */

#include "vmdsp_plugin.h"

#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>


enum{MUS_LSHORT,MUS_BYTE,MUS_LFLOAT,MUS_ERROR,MUS_NO_ERROR};
#define HAVE_JACK 1
#define MUS_LITTLE_ENDIAN 1
#define MUS_BYTE_TO_SAMPLE(n) ((float)((float)(n) / (float)(1 << 7)))
#include "vmdsp_jack_sndlib_jackpart.c"



#if 0
#include <stdio.h>
#include <time.h>
#include <unistd.h>

#include <sys/socket.h>
#endif
#include <time.h>
#include <sys/time.h>

#include <linux/soundcard.h>

struct jack_plugin {
   struct vmdsp_plugin g;
   const char* server;
   const char* app;
};

struct jack_dsp {
   struct vmdsp_dsp g;
   int speed;
   int stereo;
   int format;
   int mode;
   int started_playing;
   int dev;
};


#define container_of(ptr, type, member) ({                      \
        const typeof( ((type *)0)->member ) *__mptr = (ptr);    \
        (type *)( (char *)__mptr - offsetof(type,member) );})

#define INFO2ESD(x) container_of(x, struct jack_plugin, g)
#define ESDP(x) struct jack_plugin* jack = INFO2ESD(x)

#define DSP2ESD(x) container_of(x, struct jack_dsp, g)
#define ESDDSPP struct jack_dsp* dspp = DSP2ESD(dsp)


static int
jack_cleanup(struct vmdsp_plugin* info) {
   ESDP(info);
   //fprintf(stderr,"jack_cleanup\n");
   free(jack);
   return 0;
}



static struct vmdsp_dsp*
jack_new_dsp(struct vmdsp_plugin* info, const char* name, int mode) {
   ESDP(info);
   struct jack_dsp* dspp;

   //fprintf(stderr,"jack_new_dsp\n");

   
   dspp = calloc(1,sizeof(*dspp));
   if (dspp) {
      int fd;
      static int dasdev=10;
      memset(dspp, 0, sizeof(*dspp));
      dspp->g.info = &jack->g;

      fd = vmdsp_open64("/dev/null", mode, 0);
      if (fd < 0) {
         free(dspp);
         errno = ENODEV;
	 return NULL;
      }
      dspp->g.fd = fd;

      dspp->speed = 22050;
      dspp->stereo = 0;
      dspp->format = AFMT_U8;
      dspp->dev=dasdev++;
      dspp->mode = mode;

      return &dspp->g;
   }
   return NULL;
   (void)name;
}

static int
jack_close_dsp(struct vmdsp_dsp* dsp) {
   ESDDSPP;
   //fprintf(stderr,"jack_close_dsp\n");
   /* insert here. */
   jack_mus_audio_close(dspp->dev);
   vmdsp_close(dspp->g.fd);
   free(dspp);
   return 0;
}

static int
dspp_start_playback(struct jack_dsp* dspp) {
   //ESDP(dspp->g.info);
   //fprintf(stderr,"jack_start_playback\n");

   if (dspp->mode != O_RDWR && dspp->mode != O_WRONLY) {
      fprintf(stderr,"Would not start playback: bad mode\n");
      errno = EINVAL;
      return -1;
   }
   //dspp->wr = jack_play_stream(dspp->speed, (dspp->format == AFMT_S16_LE) ? 16 : 8, dspp->stereo ? 2 : 1, jack->app);
   jack_mus_audio_open_output(dspp->dev, dspp->speed, dspp->stereo?2:1, dspp->format==AFMT_S16_LE?MUS_LSHORT:MUS_BYTE, 1024); // size is ignored.
   return 0;
}

static int
dspp_start_record(struct jack_dsp* dspp) {
   fprintf(stderr,"jack_start_record (not supported, yet)\n");
#if 0
   ESDP(dspp->g.info);
  /* insert here */
   /* Disable recording: vmware-vmx does not cope correctly with blocking read */
   return 0;
#else
   errno = EINVAL;
   return -1;
   (void)dspp;
#endif
}

static ssize_t
dspp_read(struct vmdsp_dsp* dsp, void* buffer, size_t length) {
   //ESDDSPP;
   fprintf(stderr,"jack_read\n");
   return length;
}



static ssize_t
dspp_write(struct vmdsp_dsp* dsp, const void* buffer, size_t bytes) {
   ESDDSPP;
   //fprintf(stderr,"jack_write %d %d\n",length,(int) (1000000 * (double)length / (float)dspp->speed));

   if(!dspp->started_playing){
     if(!dspp->dev){
       return bytes;
     }
     dspp_start_playback(dspp);
     dspp->started_playing=1;
   }
   return jack_mus_audio_write(dspp->dev,(char*)buffer,bytes);
}

static int
dspp_set_speed(struct vmdsp_dsp* dsp, int speed) {
   ESDDSPP;
   //fprintf(stderr,"jack_set_speed\n");
   dspp->speed = speed;
   return 0;
}

static int
dspp_set_stereo(struct vmdsp_dsp* dsp, int stereo) {
   ESDDSPP;
   //fprintf(stderr,"jack_set_stereo\n");
   dspp->stereo = stereo;
   return 0;
}

static int
dspp_set_format(struct vmdsp_dsp* dsp, int format) {
   ESDDSPP;
   //fprintf(stderr,"jack_set_format\n");
   switch (format) {
      case AFMT_U8:
      case AFMT_S16_LE:
         break;
      default:
         format = AFMT_U8;
	 break;
   }
      
   dspp->format = format;
   return 0;
}

static int
dspp_get_format(struct vmdsp_dsp* dsp, int* format) {
   ESDDSPP;
   //fprintf(stderr,"jack_get_format\n");
   *format = dspp->format;
   return 0;
}

static const struct vmdsp_operations esd_ops = {
   .cleanup	= jack_cleanup,
   .new_dsp	= jack_new_dsp,
   .new_mixer	= NULL,
   .close_dsp	= jack_close_dsp,
   .close_mixer	= NULL,
   .read	= dspp_read,
   .write	= dspp_write,
   .set_speed	= dspp_set_speed,
   .set_stereo	= dspp_set_stereo,
   .set_format	= dspp_set_format,
   .get_format  = dspp_get_format,
   .set_volume	= NULL,
   .get_volume	= NULL,
   .set_recsrc	= NULL,
   .get_recsrc	= NULL,
};

struct vmdsp_plugin*
vmdsp_plugin_init(const char* cfgfile) {
   struct jack_plugin* esdp;
   //fprintf(stderr,"jack_plugin-init\n");

   if(jack_mus_audio_initialize()==MUS_ERROR)
     return NULL;

   esdp = malloc(sizeof(*esdp));
   if (!esdp) {
      return NULL;
   }
   memset(esdp, 0, sizeof(*esdp));
   esdp->g.d_op = &esd_ops;
   esdp->app = cfgfile;

   return &esdp->g;
}
