
/* This code is copied from sndlib and modified slightly. */


/* ------------------------------- JACK ----------------------------------------- */

/* Kjetil S. Matheussen. k.s.matheussen@notam02.no */
/* Based on code from ceres. */
 
#if HAVE_JACK
#define AUDIO_OK
#include <jack/jack.h>
#include <samplerate.h>
#include <sys/mman.h>
#include <signal.h>

#if MUS_LITTLE_ENDIAN
#  define MUS_COMP_SHORT MUS_LSHORT
#  define MUS_COMP_FLOAT MUS_LFLOAT
#else
#  define MUS_COMP_SHORT MUS_BSHORT
#  define MUS_COMP_FLOAT MUS_BFLOAT
#endif

#define SRC_QUALITY SRC_SINC_BEST_QUALITY
 
/*************/
/* Jack Part */
/*************/

#define SNDJACK_NUMINCHANNELS 2

#define SNDJACK_MAXSNDS 20

#define SNDJACK_BUFFERSIZE 32768
//#define SNDJACK_BUFFERSIZE (32768*16)

typedef jack_default_audio_sample_t sample_t;
typedef jack_nframes_t nframes_t;

struct SndjackChannel{
  jack_port_t *port;
  sample_t *buffer;
};

static jack_client_t *sndjack_client = NULL;


/*************************/
/* Variables for reading */
/*************************/
static int sndjack_num_read_channels_allocated=0;
static int sndjack_num_read_channels_inuse=0;

static struct SndjackChannel *sndjack_read_channels=NULL;

static pthread_cond_t sndjack_read_cond= PTHREAD_COND_INITIALIZER;

static pthread_mutex_t sndjack_read_mutex= PTHREAD_MUTEX_INITIALIZER;
static int sj_r_buffersize=0;
static int sj_r_writeplace=0;
static int sj_r_readplace=0;
static int sj_r_unread=0;
static int sj_r_xrun=0;
static int sj_r_totalxrun=0;

/*************************/
/* Variables for writing */
/*************************/
static pthread_cond_t sndjack_cond= PTHREAD_COND_INITIALIZER;
static pthread_mutex_t sndjack_mutex= PTHREAD_MUTEX_INITIALIZER;

enum{SJ_STOPPED,SJ_RUNNING,SJ_ABOUTTOSTOP};

// Variables for the ringbuffer:
static  int sj_writeplace=0;
static  int sj_readplace=0;
static  int sj_unread=0;
static  int sj_buffersize;
static int sj_jackbuffersize; // number of frames sent to sndjack_process.
static int sj_totalxrun=0;
static int sj_xrun=0;
static int sj_status=SJ_STOPPED;

static int sndjack_num_channels_allocated=0;
static int sndjack_num_channels_inuse=0;
static struct SndjackChannel *sndjack_channels=NULL;
static int sndjack_read_format;

static SRC_STATE **sndjack_srcstates;
static double sndjack_srcratio=1.0;

static int jack_mus_watchdog_counter=0;


#define SJ_MAX(a,b) (((a)>(b))?(a):(b))
#define SJ_MIN(a,b) (((a)<(b))?(a):(b))

static void sndjack_read_process(jack_nframes_t nframes){
  int i,ch;
  sample_t *out[sndjack_num_channels_allocated];

  if(sndjack_num_read_channels_inuse==0) return;

  for(ch=0;ch<sndjack_num_read_channels_allocated;ch++){
    out[ch]=(sample_t*)jack_port_get_buffer(sndjack_read_channels[ch].port,nframes);
  }

  for(i=0;i<(int)nframes;i++){
    if(sj_r_unread==sj_buffersize){
      sj_r_xrun+=nframes-i;
      goto exit;
    }
    for(ch=0;ch<sndjack_num_read_channels_inuse;ch++)
      sndjack_read_channels[ch].buffer[sj_r_writeplace]=out[ch][i];
    sj_r_unread++;
    sj_r_writeplace++;
    if(sj_r_writeplace==sj_r_buffersize)
      sj_r_writeplace=0;
  }
 exit:
  pthread_cond_broadcast(&sndjack_read_cond);
}


static void sndjack_write_process(jack_nframes_t nframes){
  int ch,i;
  sample_t *out[sndjack_num_channels_allocated];

  for(ch=0;ch<sndjack_num_channels_allocated;ch++){
    out[ch]=(sample_t*)jack_port_get_buffer(sndjack_channels[ch].port,nframes);
  }

  if(sj_status==SJ_STOPPED){
    for(ch=0;ch<sndjack_num_channels_allocated;ch++){
      memset(out[ch],0,nframes*sizeof(sample_t));
    }
  }else{

    // First null out unused channels, if any.
    if(sndjack_num_channels_inuse==1 && sndjack_num_channels_allocated>=2){
      for(ch=2;ch<sndjack_num_channels_allocated;ch++){
	memset(out[ch],0,nframes*sizeof(sample_t));
      }
    }else{
      for(ch=sndjack_num_channels_inuse;ch<sndjack_num_channels_allocated;ch++){
	memset(out[ch],0,nframes*sizeof(sample_t));
      }
    }

    for(i=0;i<(int)nframes;i++){
      if(sj_unread==0){	
	if(sj_status==SJ_RUNNING){
	  sj_xrun+=nframes-i;
	  //printf("XRUN! %d\n",nframes-i);
	}
	for(;i<(int)nframes;i++){
	  for(ch=0;ch<sndjack_num_channels_inuse;ch++){
	    out[ch][i]=0.0f;
	  }
	}
	break;
      }

      if(sndjack_num_channels_inuse==1 && sndjack_num_channels_allocated>=2){
	for(ch=0;ch<2;ch++){
	  out[ch][i]=sndjack_channels[0].buffer[sj_readplace];
	}
      }else{
	for(ch=0;ch<sndjack_num_channels_inuse;ch++){
	  out[ch][i]=sndjack_channels[ch].buffer[sj_readplace];
	}
      }
      sj_unread--;
      sj_readplace++;
      if(sj_readplace==sj_buffersize)
	sj_readplace=0;
    }
    
    pthread_cond_broadcast(&sndjack_cond);

    if(sj_status==SJ_ABOUTTOSTOP && sj_unread==0)
      sj_status=SJ_STOPPED;
  }

}
 


static int sndjack_process(jack_nframes_t nframes, void *arg){
  sndjack_read_process(nframes);
  sndjack_write_process(nframes);
  return 0;
}


static int sndjack_read(void *buf,int bytes,int chs){
  int i,ch;
  int nframes=bytes /
    sndjack_read_format==MUS_COMP_FLOAT ? sizeof(float) :
    sndjack_read_format==MUS_COMP_SHORT ? sizeof(short) :
    1;
  float *buf_f=buf;
  short *buf_s=buf;
  char *buf_c=buf;

  for(i=0;i<nframes;i++){
    while(sj_r_unread==0){
      pthread_cond_wait(&sndjack_read_cond,&sndjack_read_mutex);
      jack_mus_watchdog_counter++;
    }

    if(sj_r_xrun>0){
      sj_r_totalxrun+=sj_r_xrun;
      sj_r_xrun=0;
      return -1;
    }
    for(ch=0;ch<chs;ch++){
      switch(sndjack_read_format){
      case MUS_BYTE:
	buf_c[i*chs+ch]=sndjack_read_channels[ch].buffer[sj_r_readplace] * 127.9f;
	break;
      case MUS_COMP_SHORT:
	buf_s[i*chs+ch]=sndjack_read_channels[ch].buffer[sj_r_readplace] * 32767.9f;
	break;
      case MUS_COMP_FLOAT:
	buf_f[i*chs+ch]=sndjack_read_channels[ch].buffer[sj_r_readplace];
	break;
      }}
    sj_r_unread--;
    sj_r_readplace++;
    if(sj_r_readplace==sj_r_buffersize)
      sj_r_readplace=0;
  }
  return 0;
}

static void sndjack_write(sample_t **buf,int nframes,int latencyframes,int chs){
  int ch;
  int i;
  int have_waited=0;

  if(sj_xrun>0){
    if(sj_status==SJ_RUNNING){
      printf("Warning. %d frames delayed.\n",sj_xrun);
      sj_totalxrun+=sj_xrun;
    }
    sj_xrun=0;
  }

  for(i=0;i<nframes;i++){
    while(
	  sj_unread==sj_buffersize
	  //|| sj_unread >= SJ_MAX(sj_jackbuffersize*2, latencyframes)
	  )
      {
	jack_mus_watchdog_counter++;
	pthread_cond_wait(&sndjack_cond,&sndjack_mutex);
	have_waited=1;
      }

    for(ch=0;ch<chs;ch++)
      sndjack_channels[ch].buffer[sj_writeplace]=buf[ch][i];
    
    sj_unread++;
    sj_writeplace++;
    if(sj_writeplace==sj_buffersize)
      sj_writeplace=0;
  }

  jack_mus_watchdog_counter++;

  if(sj_status!=SJ_RUNNING){
    if(sj_unread>=SJ_MAX(1024,sj_jackbuffersize))
      sj_status=SJ_RUNNING;
  }

}
 
static int sndjack_buffersizecallback(jack_nframes_t nframes, void *arg){
  sj_jackbuffersize=nframes;
  return 0;
}

static int sndjack_getnumoutchannels(void){
  int lokke=0;
  const char **ports=jack_get_ports(sndjack_client,NULL,NULL,JackPortIsPhysical|JackPortIsInput);
  while(ports!=NULL && ports[lokke]!=NULL){
    lokke++;
  }
  if(lokke<2) return 2;
  return lokke;
}

static int sndjack_getnuminchannels(void){
  int lokke=0;
  const char **ports=jack_get_ports(sndjack_client,NULL,NULL,JackPortIsPhysical|JackPortIsOutput);
  while(ports!=NULL && ports[lokke]!=NULL){
    lokke++;
  }
  if(lokke<2) return 2;
  return lokke;
}


static int sndjack_init(void){
  int ch;
  int numch;
  int numch_read;
  int num=0;

  while(num<SNDJACK_MAXSNDS){
    char temp[500];
    sprintf(temp,"vmwaredsp%d",num);
    if ((sndjack_client=jack_client_new(temp)) != 0) {
      break;
    }
    num++;
  }

  if(sndjack_client==NULL){
    fprintf(stderr,"Unable to create new jack_client\n");
    return -1;
  }

  pthread_mutex_init(&sndjack_mutex,NULL);
  pthread_cond_init(&sndjack_cond,NULL);
  pthread_mutex_init(&sndjack_read_mutex,NULL);
  pthread_cond_init(&sndjack_read_cond,NULL);

  jack_set_process_callback(sndjack_client,sndjack_process,NULL);

  sndjack_num_channels_allocated = numch = sndjack_getnumoutchannels();
  numch_read=sndjack_getnuminchannels();
  sndjack_num_read_channels_allocated=SJ_MAX(SNDJACK_NUMINCHANNELS,numch_read);

  sndjack_channels=calloc(sizeof(struct SndjackChannel),numch);
  sndjack_read_channels=calloc(sizeof(struct SndjackChannel),sndjack_num_read_channels_allocated);

  for(ch=0;ch<numch;ch++){
    sndjack_channels[ch].buffer=calloc(sizeof(sample_t),SNDJACK_BUFFERSIZE);
  }
  for(ch=0;ch<sndjack_num_read_channels_allocated;ch++){
    sndjack_read_channels[ch].buffer=calloc(sizeof(sample_t),SNDJACK_BUFFERSIZE);
  }
  sj_buffersize=SNDJACK_BUFFERSIZE;

  for(ch=0;ch<numch;ch++){
    char temp[500];
    sprintf(temp,"out_%d",ch+1);
    if((sndjack_channels[ch].port=jack_port_register(
						     sndjack_client,
						     strdup(temp),
						     JACK_DEFAULT_AUDIO_TYPE,
						     JackPortIsOutput,
						     0
						     ))==NULL)
      {
	fprintf(stderr,"Error. Could not register jack port.\n");
	goto failed_register;
      }
  }

  for(ch=0;ch<sndjack_num_read_channels_allocated;ch++){
    char temp[500];
    sprintf(temp,"in_%d",ch+1);
    if((sndjack_read_channels[ch].port=jack_port_register(
							  sndjack_client,
							  strdup(temp),
							  JACK_DEFAULT_AUDIO_TYPE,
							  JackPortIsInput,
							  0
							  ))==NULL)
      {
	fprintf(stderr,"Error. Could not register jack port.\n");
	goto failed_register;
      }
  }




  sj_jackbuffersize=jack_get_buffer_size(sndjack_client);
  jack_set_buffer_size_callback(sndjack_client,sndjack_buffersizecallback,NULL);

  if (jack_activate (sndjack_client)) {
    fprintf (stderr, "Error. Cannot activate jack client.\n");
    goto failed_activate;
  }

  {
    const char **outportnames=jack_get_ports(sndjack_client,NULL,NULL,JackPortIsPhysical|JackPortIsInput);
    for(ch=0;outportnames && outportnames[ch]!=NULL && ch<numch;ch++){
      if (
	  jack_connect(
		       sndjack_client,
		       jack_port_name(sndjack_channels[ch].port),
		       outportnames[ch]
		       )
	  )
	{
	  printf ("Warning. Cannot connect jack output port %d: \"%s\".\n",ch,outportnames[ch]);
	}
    }
  }

  {    
    const char **inportnames=jack_get_ports(sndjack_client,NULL,NULL,JackPortIsPhysical|JackPortIsOutput);
    for(ch=0;inportnames && inportnames[ch]!=NULL && ch<numch;ch++){
    if (
	jack_connect(
		     sndjack_client,
		     inportnames[ch],
		     jack_port_name(sndjack_read_channels[ch].port)
		     )
	)
      {
	printf ("Warning. Cannot connect jack input port %d: \"%s\".\n",ch,inportnames[ch]);
      }
    }
  }

  munlockall(); // jack calls mlockall, but vmware is too large to do that:  "vcpu-0| Could not mmap paging file : Resource temporarily unavailable"

  // Instead we just do this: (which is not enough, but maybe better than nothing)
  {
    mlock(sndjack_channels,sizeof(struct SndjackChannel)*sndjack_num_channels_allocated);
    mlock(sndjack_read_channels,sizeof(struct SndjackChannel)*sndjack_num_read_channels_allocated);

    for(ch=0;ch<numch;ch++){
      mlock(sndjack_channels[ch].buffer,sizeof(sample_t)*SNDJACK_BUFFERSIZE);
    }
    for(ch=0;ch<sndjack_num_read_channels_allocated;ch++){
      mlock(sndjack_read_channels[ch].buffer,sizeof(sample_t)*SNDJACK_BUFFERSIZE);
    }
  }
  
  return 0;
  
  // failed_connect:
 failed_activate:
  jack_deactivate(sndjack_client);
  
 failed_register:
  jack_client_close(sndjack_client);
  sndjack_client=NULL;

  return -1;
}
static void sndjack_cleanup(void){
  int ch;
  for(ch=0;ch<sndjack_num_channels_allocated;ch++){
    src_delete(sndjack_srcstates[ch]);
  }
  jack_deactivate(sndjack_client);
  jack_client_close(sndjack_client);

}



/***************/
/* Sndlib Part */
/***************/

static int sndjack_format;
static sample_t **sndjack_buffer;
static sample_t *sndjack_srcbuffer;

static int sndjack_dev;
static int sndjack_read_dev;

/* prototypes for the jack sndlib functions */
static int   jack_mus_audio_initialize(void);
#if 0
static void  jack_mus_oss_set_buffers(int num, int size);
static int   jack_mus_audio_systems(void);
static char* jack_mus_audio_system_name(int system);
static char* jack_mus_audio_moniker(void);
#endif
static int   jack_mus_audio_open_output(int ur_dev, int srate, int chans, int format, int size);
static int   jack_mus_audio_open_input(int ur_dev, int srate, int chans, int format, int requested_size);
static int   jack_mus_audio_write(int id, char *buf, int bytes);
static int   jack_mus_audio_read(int id, char *buf, int bytes);
static int   jack_mus_audio_close(int id);
#if 0
static int   jack_mus_audio_mixer_read(int ur_dev, int field, int chan, float *val);
static int   jack_mus_audio_mixer_write(int ur_dev, int field, int chan, float *val);
static void  jack_describe_audio_state_1(void);
#endif

static int jack_mus_audio_initialize(void) {
  int ch;
  static int audio_initialized=0;

  if(audio_initialized){
    return MUS_NO_ERROR;
  }


  if(sndjack_init()!=0)
    return MUS_ERROR;

  sndjack_buffer=calloc(sizeof(sample_t*),sndjack_num_channels_allocated);
  for(ch=0;ch<sndjack_num_channels_allocated;ch++)
    sndjack_buffer[ch]=calloc(sizeof(sample_t),SNDJACK_BUFFERSIZE);
  sndjack_srcbuffer=calloc(sizeof(sample_t),SNDJACK_BUFFERSIZE);

  sndjack_srcstates=calloc(sizeof(SRC_STATE*),sndjack_num_channels_allocated);
  for(ch=0;ch<sndjack_num_channels_allocated;ch++){
    sndjack_srcstates[ch]=src_new(SRC_QUALITY,1,NULL);
  }

  atexit(sndjack_cleanup);

  /*
  api = JACK_API;
  vect_mus_audio_initialize = jack_mus_audio_initialize;
  vect_mus_oss_set_buffers = jack_mus_oss_set_buffers;
  vect_mus_audio_systems = jack_mus_audio_systems;
  vect_mus_audio_system_name = jack_mus_audio_system_name;
  vect_mus_audio_moniker = jack_mus_audio_moniker;
  vect_mus_audio_open_output = jack_mus_audio_open_output;
  vect_mus_audio_open_input = jack_mus_audio_open_input;
  vect_mus_audio_write = jack_mus_audio_write;
  vect_mus_audio_read = jack_mus_audio_read;
  vect_mus_audio_close = jack_mus_audio_close;
  vect_mus_audio_mixer_read = jack_mus_audio_mixer_read;
  vect_mus_audio_mixer_write = jack_mus_audio_mixer_write;
  vect_describe_audio_state_1 = jack_describe_audio_state_1;

  */

  audio_initialized = 1;


  /* Locking all future memory shouldn't be that necessary, and might even freeze the machine in certain situations. */
  /* So remove MCL_FUTURE from the mlockall call. (No. We can't do that. It can screw up code using the realtime extension. -Kjetil.*/
  //munlockall();
  //mlockall(MCL_CURRENT);
  
  return MUS_NO_ERROR;
}

static int jack_mus_isrunning=0;
static pid_t jack_mus_player_pid;  
static pthread_t jack_mus_watchdog_thread;

static void *jack_mus_audio_watchdog(void *arg){
  struct sched_param par;

  par.sched_priority = sched_get_priority_max(SCHED_RR);
  if(sched_setscheduler(0,SCHED_RR,&par)==-1){
    fprintf(stderr,"SNDLIB: Unable to set SCHED_RR realtime priority for the watchdog thread. No watchdog.\n");
    goto exit;
  }

  for(;;){
    int last=jack_mus_watchdog_counter;
    sleep(1);

    if(jack_mus_isrunning && jack_mus_watchdog_counter<last+10){
      struct sched_param par;
      fprintf(stderr,"SNDLIB: Setting player to non-realtime for 2 seconds.\n");

      par.sched_priority = 0;
      if(sched_setscheduler(jack_mus_player_pid,SCHED_OTHER,&par)==-1){
	fprintf(stderr,"SNDLIB: Unable to set non-realtime priority. Must kill player thread. Sorry!\n");
	while(1){
	  kill(jack_mus_player_pid,SIGKILL);
	  sleep(2);
	}
      }

      sleep(2);

      if(jack_mus_isrunning){
	par.sched_priority = sched_get_priority_min(SCHED_RR)+1;
	if(sched_setscheduler(jack_mus_player_pid,SCHED_RR,&par)==-1){
	  fprintf(stderr,"SNDLIB: Could not set back to realtime priority...\n");
	}else
	  fprintf(stderr,"SNDLIB: Play thread set back to realtime priority.\n");
      }

    }
  }
 exit:
  fprintf(stderr,"SNDLIB: Watchdog exiting\n");
  return NULL;
}



static void jack_mus_audio_set_realtime(void){
  struct sched_param par;
  static int watchdog_started=0;

  jack_mus_player_pid=getpid();

  if(watchdog_started==0){
    if(pthread_create(&jack_mus_watchdog_thread,NULL,jack_mus_audio_watchdog,NULL)!=0){
      fprintf(stderr,"Could not create watchdog. Not running realtime\n");
      return;
    }
    watchdog_started=1;
  }

  jack_mus_isrunning=1;

  par.sched_priority = sched_get_priority_min(SCHED_RR)+1;
  if(sched_setscheduler(0,SCHED_RR,&par)==-1){
    fprintf(stderr,"SNDLIB: Unable to set SCHED_RR realtime priority for the player thread.\n");
  }{
    //fprintf(stderr,"Set realtime priority\n");
  }
}

static void jack_mus_audio_set_non_realtime(void){
  struct sched_param par;
  par.sched_priority = 0;
  sched_setscheduler(0,SCHED_OTHER,&par);
  //fprintf(stderr,"Set non-realtime priority\n");
  jack_mus_isrunning=0;
}

static int jack_mus_audio_open_output(int dev, int srate, int chans, int format, int size){
  if(sndjack_client==NULL){
    if(jack_mus_audio_initialize()==MUS_ERROR)
      return MUS_ERROR;
  }
  
  if(sndjack_num_channels_allocated<chans){
    printf("Error. Can not play back %d channels. (Only %d)\n",chans,sndjack_num_channels_allocated);
    return MUS_ERROR;
  }

  if(format!=MUS_BYTE && format!=MUS_COMP_SHORT && format!=MUS_COMP_FLOAT){
    //printf("Error, unable to handle format %s.\n",mus_data_format_to_string(format));
    printf("Error, unable to handle format %s.\n","(not available)");
    return MUS_ERROR;
  }

  while(sj_status!=SJ_STOPPED) usleep(5);

  sj_unread=0;
  sj_writeplace=0;
  sj_readplace=0;

  if(srate!=(int)jack_get_sample_rate(sndjack_client)){
    int lokke;
    //printf("Warning, sample-rate differs between snd and jack. Sound will not be played correctly! %d/%d\n",srate,jack_get_sample_rate(sndjack_client));
    sndjack_srcratio=(double)jack_get_sample_rate(sndjack_client)/(double)srate;
    for(lokke=0;lokke<chans;lokke++){
      src_reset(sndjack_srcstates[lokke]);
    }
  }else{
    sndjack_srcratio=1.0;
  }

  sndjack_format=format;
  sndjack_num_channels_inuse=chans;
  sndjack_dev=dev;

  //jack_mus_audio_set_realtime(); //commented out for vmdsp.

  return(MUS_NO_ERROR);
}
 
static int sndjack_from_byte(int ch,int chs,char *buf,float *out,int bytes){
  int i;
  int len=bytes/chs;
  if(len>SNDJACK_BUFFERSIZE) return -1;

  for(i=0;i<len;i++){
    out[i]=MUS_BYTE_TO_SAMPLE(buf[i*chs+ch]);
  }
  return len;
}

static int sndjack_from_short(int ch,int chs,short *buf,float *out,int bytes){
  int i;
  int len=bytes/(sizeof(short)*chs);
  if(len>SNDJACK_BUFFERSIZE) return -1;

  for(i=0;i<len;i++){
    out[i]=(float)buf[i*chs+ch]/32768.1f;
  }
  return len;
}

static int sndjack_from_float(int ch,int chs,float *buf,float *out,int bytes){
  int i;
  int len=bytes/(sizeof(float)*chs);
  if(len>SNDJACK_BUFFERSIZE) return -1;

  for(i=0;i<len;i++){
    out[i]=buf[i*chs+ch];
  }
  return len;
}


static int jack_mus_audio_write(int line, char *buf, int bytes){
  int ch;
  int outlen=0;

  {
    int bytes_per_frame=sndjack_num_channels_inuse * (sndjack_format==MUS_BYTE?sizeof(float):sizeof(short));

    int max_frames=SJ_MAX(sj_jackbuffersize,1024)*2;
    int min_frames=1; // If we send out 0 frames, "/dev/dsp" is closed.
    int num_frames=bytes/bytes_per_frame;

    if(num_frames>max_frames*3/2 && sj_status==SJ_RUNNING){
      pthread_cond_wait(&sndjack_cond,&sndjack_mutex);
    }

    int out_frames=max_frames-sj_unread;
    
    if(out_frames>num_frames){
      out_frames=num_frames;
    }else{
      if(out_frames<min_frames)
	out_frames=SJ_MIN(num_frames,min_frames);
    }
    bytes=bytes_per_frame * out_frames;
  }

  for(ch=0;ch<sndjack_num_channels_inuse;ch++){
    int len = 0;
    float *buf2=sndjack_srcratio==1.0?sndjack_buffer[ch]:sndjack_srcbuffer;

    switch(sndjack_format){
    case MUS_BYTE:
      len=sndjack_from_byte(ch,sndjack_num_channels_inuse,buf,buf2,bytes);
      break;
    case MUS_COMP_SHORT:
      len=sndjack_from_short(ch,sndjack_num_channels_inuse,(short *)buf,buf2,bytes);
      break;
    case MUS_COMP_FLOAT:
      len=sndjack_from_float(ch,sndjack_num_channels_inuse,(float *)buf,buf2,bytes);
      break;
    }
    if(len<0){
      printf("Errur. Input buffer to large for mus_audio_write.\n");
      return 0;
    }

    if(sndjack_srcratio!=1.0){
      SRC_DATA src_data={
	buf2,sndjack_buffer[ch],
	len,SNDJACK_BUFFERSIZE,
	0,0,
	0,
	sndjack_srcratio
      };
      int res=src_process(sndjack_srcstates[ch],&src_data);
      if(res!=0){
	printf("Error while resampling. (%s)\n",src_strerror(res));
	return 0;
      }
      if(src_data.input_frames!=len){
	printf("Unsuccessfull resampling: Should have used %d bytes, used %d.",(int)len,(int)src_data.input_frames);
	return 0;
      }
      if(ch>0 && src_data.output_frames_gen!=outlen){
	printf("Error, src_process did not output the same number of frames as previous resampled channel (%d/%d).\n"
	       "Please report this problem to k.s.matheussen@notam02.no. Thanks!\n",(int)src_data.output_frames_gen,(int)outlen);
	return 0;
      }
      outlen=src_data.output_frames_gen;
    }else{
      outlen=len;
    }
  }


  sndjack_write(sndjack_buffer,outlen,outlen*2,sndjack_num_channels_inuse);

  //  return MUS_NO_ERROR;

  // printf("Got %d in, put out %d, max: %d, out_frames: %d, sj_unread: %d, left: %d\n",bytesorg/bytes_per_frame,bytes/bytes_per_frame,sj_jackbuffersize*2,out_frames,sj_unread,sj_jackbuffersize*2-sj_unread);
  return bytes;
}
 
static int jack_mus_audio_close(int line) 
{
  //jack_mus_audio_set_non_realtime(); // commented out for vmdsp
  if(line==sndjack_dev){
    sj_status=SJ_ABOUTTOSTOP;
    sndjack_num_channels_inuse=0;
  }
  return MUS_NO_ERROR;
 }

#if 0
int jack_mus_audio_mixer_read(int dev, int field, int chan, float *val) 
{
  //printf("dev: %d, field: %d, chan: %d\n",dev,field,chan);

  switch(field){
  case MUS_AUDIO_FORMAT:
    val[1]=MUS_COMP_FLOAT;
    val[0]=1;
    break;
  case MUS_AUDIO_PORT:
    val[0]=1;
    val[1]=MUS_AUDIO_DIGITAL_IN;
    break;
  case MUS_AUDIO_CHANNEL:
    val[0]=sndjack_num_read_channels_allocated;
    break;
  case MUS_AUDIO_AMP: 
    val[0] = 1.0f;
    break;
  default:
    printf("Got unknown request with field %d %d\n",field, MUS_AUDIO_AMP);
    return MUS_ERROR;
  }

  return MUS_NO_ERROR;
}

int jack_mus_audio_mixer_write(int dev, int field, int chan, float *val){
  return(MUS_NO_ERROR);
}
#endif

static int jack_mus_audio_open_input(int dev, int srate, int chans, int format, int size){
  if(sndjack_client==NULL){
    if(jack_mus_audio_initialize()==MUS_ERROR)
      return MUS_ERROR;
  }
  
  if(sndjack_num_read_channels_allocated<chans){
    printf("Error. Can not record %d channels. (Only %d)\n",chans,sndjack_num_read_channels_allocated);
    return MUS_ERROR;
  }

  printf("dev: %d\n" ,dev);
  if(format!=MUS_BYTE && format!=MUS_COMP_SHORT && format!=MUS_COMP_FLOAT){
    //printf("Error, unable to handle format %s.\n",mus_data_format_to_string(format));
    printf("Error, unable to handle format %s.\n","(not available)");
    return MUS_ERROR;
  }

  if(srate!=(int)jack_get_sample_rate(sndjack_client)){
    printf("Warning, jacks samplerate is %d (and not %d), and the recording will use this samplerate too.\n",jack_get_sample_rate(sndjack_client),srate);
  }

  sndjack_read_format=format;
  sndjack_num_read_channels_inuse=chans;
  sndjack_read_dev=dev;

  return(MUS_NO_ERROR);
}

static int jack_mus_audio_read(int line, char *buf, int bytes){
  if(sndjack_read(buf,bytes,sndjack_num_read_channels_inuse)==-1)
    return(MUS_ERROR);
  return MUS_NO_ERROR;
}

#if 0
static void jack_describe_audio_state_1(void) {
  char temp[500];

  pprint("jack audio:\n");
  sprintf(temp,"\tNumber of output channels: %d\n",sndjack_num_channels_allocated);pprint(temp);
  sprintf(temp,"\tNumber of input channels: %d\n",sndjack_num_read_channels_allocated);pprint(temp);
  sprintf(temp,"\tSamplerate: %d\n",jack_get_sample_rate(sndjack_client));pprint(temp);
  sprintf(temp,"\tJack buffersize: %d\n",sj_jackbuffersize);pprint(temp);
  sprintf(temp,"\tSndjack buffersize: %d\n",SNDJACK_BUFFERSIZE);pprint(temp);
  sprintf(temp,"\tMax number of instances: %d\n",SNDJACK_MAXSNDS);pprint(temp);
  sprintf(temp,"\tTotal number of frames delayed: %d\n",sj_totalxrun);pprint(temp);
  sprintf(temp,"\tCurrent cpu-load: %f\n",jack_cpu_load(sndjack_client));pprint(temp);
  sprintf(temp,"\tIs running realtime: %s\n",jack_is_realtime(sndjack_client)==1?"yes":"no");pprint(temp);
  sprintf(temp,"\tResample quality (only used when needed): %s (%s)\n",src_get_name(SRC_QUALITY),src_get_description(SRC_QUALITY));pprint(temp);
  sprintf(temp,"\tIs able to handle the following audio formats: %s %s %s\n",mus_data_format_to_string(MUS_BYTE),mus_data_format_to_string(MUS_COMP_SHORT),mus_data_format_to_string(MUS_COMP_FLOAT));pprint(temp);
  sprintf(temp,"\tPrefered audio format: %s\n",mus_data_format_to_string(MUS_COMP_FLOAT));pprint(temp);
}


int jack_mus_audio_systems(void) {
  return(2);
}

char *jack_mus_audio_system_name(int system) {return("linux jack");}
char *jack_mus_audio_moniker(void) {return("jack");}
#endif

#endif
